#define GL_GLEXT_PROTOTYPES
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/extensions/Xcomposite.h>
#include <X11/extensions/Xdamage.h>
#include <X11/extensions/shape.h>
#include <X11/extensions/Xrandr.h>
#include <GL/gl.h>
#include <GL/glx.h>

typedef struct CompositedWindow
{
	Window window;
	int x;
	int y;
	unsigned width;
	unsigned height;
	unsigned border;
	int map_state;
	Damage damage;
	Pixmap pixmap;
	GLXPixmap glx_pixmap;
	int recreate_pixmap;
	unsigned texture;
	unsigned mask_texture;
	int use_mask;
	int recreate_mask;
} CompositedWindow;

enum
{
	CYLINDRICAL = 1,
	SPHERICAL
};

typedef struct CompositedMonitor
{
	char *name;
	int enabled;
	int x;
	int y;
	unsigned width;
	unsigned height;
	float keystone_vertical;
	float curvature_type;
	float curvature_depth;
	float vertical_center;
	float perspective;
	unsigned vertex_buffer;
	unsigned index_buffer;
	unsigned vertex_array;
	unsigned tessellation;
	unsigned nelements;
	short *geometry_data;
	unsigned geometry_data_size;
} CompositedMonitor;

typedef struct CompositedScreen
{ 
	int number;
	Window root;
	unsigned width;
	unsigned height;
	Window overlay;
	Window render_window;
	GLXFBConfig fbconfig;
	GLXWindow glx_window;
	GLXContext glx_context;
	unsigned shaders[3];
	unsigned program;
	int geometry_loc;
	unsigned masked_program;
	int masked_geometry_loc;
	unsigned window_vertex_buffer;
	unsigned window_vertex_array;
	unsigned framebuffer;
	unsigned fb_texture;
	Pixmap root_pixmap;
	GLXPixmap root_glx_pixmap;
	unsigned root_texture;
	CompositedWindow *windows;
	unsigned nwindows;
	unsigned windows_capacity;
	CompositedMonitor *monitors;
	unsigned nmonitors;
	int dirty;
} CompositedScreen;

typedef struct Compositor
{
	Display *display;
	CompositedScreen *screens;
	unsigned nscreens;
	int damage_event;
	int shape_event;
	int randr_event;
	PFNGLXCREATECONTEXTATTRIBSARBPROC glXCreateContextAttribsARB;
	PFNGLXBINDTEXIMAGEEXTPROC glXBindTexImageEXT;
	PFNGLXRELEASETEXIMAGEEXTPROC glXReleaseTexImageEXT;
	Atom root_pmap_atom;
	Atom correction_atom;
	Atom monitors_atom;
	Atom geometry_data_atom;
	Window selection_window;
	int dirty;
} Compositor;

static const char *vshader_src =
	"#version 130\n"
	"uniform vec4 geometry;\n"
	"in vec2 vertex;\n"
	"in vec2 texture_coord;\n"
	"out vec2 texcoord;\n"
	"void main()\n"
	"{\n"
	"  gl_Position = vec4((geometry.xy+vertex*geometry.zw)*2.0-1.0, 0.0, 1.0);\n"
	"  texcoord = texture_coord;\n"
	"}\n";

static const char *fshader_src =
	"#version 130\n"
	"uniform sampler2D image;\n"
	"in vec2 texcoord;\n"
	"out vec4 frag_color;\n"
	"void main()\n"
	"{\n"
	"  frag_color = texture(image, texcoord);\n"
	"}\n";

static const char *masked_fshader_src =
	"#version 130\n"
	"uniform sampler2D image;\n"
	"uniform sampler2D mask;\n"
	"in vec2 texcoord;\n"
	"out vec4 frag_color;\n"
	"void main()\n"
	"{\n"
	"  if(texture(mask, texcoord).r==0.0)\n"
	"    discard;\n"
	"  frag_color = texture(image, texcoord);\n"
	"}\n";

static const float window_vertices[] =
{
	/* vertex    texcoord */
	0.0f, 1.0f,  0.0f, 0.0f,
	0.0f, 0.0f,  0.0f, 1.0f,
	1.0f, 1.0f,  1.0f, 0.0f,
	1.0f, 0.0f,  1.0f, 1.0f
};

int terminate_requested = 0;

int x_error_handler(Display *display, XErrorEvent *event)
{
	printf("Ignoring X error %d on resource %lx\n", event->error_code, event->resourceid);
	(void)display;
	return 0;
}

int with_error(const char *message)
{
	fprintf(stderr, "%s\n", message);
	return 0;
}

int initialize_gl(Compositor *compositor, CompositedScreen *screen)
{
	int attribs[9];
	unsigned i;
	GLXFBConfig *configs;
	int nconfigs;
	XVisualInfo *vi;
	XSetWindowAttributes win_attr;

	i = 0;
	attribs[i++] = GLX_DRAWABLE_TYPE;
	attribs[i++] = GLX_WINDOW_BIT;
	attribs[i++] = GLX_RENDER_TYPE;
	attribs[i++] = GLX_RGBA_BIT;
	attribs[i++] = GLX_DOUBLEBUFFER;
	attribs[i++] = True;
	attribs[i++] = GLX_BIND_TO_TEXTURE_RGBA_EXT;
	attribs[i++] = True;
	attribs[i] = None;

	configs = glXChooseFBConfig(compositor->display, screen->number, attribs, &nconfigs);
	if(!configs || !nconfigs)
		return with_error("Could not find a suitable FBConfig");
	screen->fbconfig = configs[0];
	XFree(configs);

	vi = glXGetVisualFromFBConfig(compositor->display, screen->fbconfig);
	win_attr.colormap = XCreateColormap(compositor->display, screen->root, vi->visual, AllocNone);
	screen->render_window = XCreateWindow(compositor->display, screen->overlay, 0, 0, screen->width, screen->height, 0, vi->depth, InputOutput, vi->visual, CWColormap, &win_attr);
	XMapWindow(compositor->display, screen->render_window);

	screen->glx_window = glXCreateWindow(compositor->display, screen->fbconfig, screen->render_window, NULL);

	i = 0;
	attribs[i++] = GLX_CONTEXT_FLAGS_ARB;
	attribs[i++] = GLX_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB;
	attribs[i++] = GLX_CONTEXT_MAJOR_VERSION_ARB;
	attribs[i++] = 3;
	attribs[i++] = GLX_CONTEXT_MINOR_VERSION_ARB;
	attribs[i++] = 1;
	attribs[i] = None;

	screen->glx_context = compositor->glXCreateContextAttribsARB(compositor->display, screen->fbconfig, NULL, True, attribs);

	XFree(vi);

	return 1;
}

void use_gl(Compositor *compositor, CompositedScreen *screen)
{
	glXMakeContextCurrent(compositor->display, screen->glx_window, screen->glx_window, screen->glx_context);
}

unsigned compile_shader(GLenum type, const char *source)
{
	unsigned shader;
	int status;
	char info_log[1024];
	GLsizei length;
	
	shader = glCreateShader(type);
	glShaderSource(shader, 1, &source, NULL);
	glCompileShader(shader);
	glGetShaderiv(shader, GL_COMPILE_STATUS, &status);
	glGetShaderInfoLog(shader, sizeof(info_log), &length, info_log);
	if(!status)
	{
		fprintf(stderr, "Shader compilation failed:\n%s\n", info_log);
		glDeleteShader(shader);
		return 0;
	}
	else if(length)
		printf("Shader info log:\n%s\n", info_log);

	return shader;
}

unsigned link_program(unsigned vshader, unsigned fshader)
{
	unsigned program;
	int status;
	char info_log[1024];
	GLsizei length;

	program = glCreateProgram();
	glAttachShader(program, vshader);
	glAttachShader(program, fshader);
	glBindAttribLocation(program, 0, "vertex");
	glBindAttribLocation(program, 1, "texture_coord");
	glBindFragDataLocation(program, 0, "frag_color");
	glLinkProgram(program);

	glGetProgramiv(program, GL_LINK_STATUS, &status);
	glGetProgramInfoLog(program, sizeof(info_log), &length, info_log);
	if(!status)
	{
		fprintf(stderr, "Program link failed:\n%s\n", info_log);
		glDeleteProgram(program);
		return 0;
	}
	else if(length)
		printf("Program info log:\n%s\n", info_log);

	return program;
}

unsigned create_2d_texture()
{
	unsigned texture;
	glGenTextures(1, &texture);
	glBindTexture(GL_TEXTURE_2D, texture);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	return texture;
}

int create_gl_resources(CompositedScreen *screen)
{
	unsigned stride;
	int loc;

	screen->shaders[0] = compile_shader(GL_VERTEX_SHADER, vshader_src);
	screen->shaders[1] = compile_shader(GL_FRAGMENT_SHADER, fshader_src);
	screen->shaders[2] = compile_shader(GL_FRAGMENT_SHADER, masked_fshader_src);
	if(!screen->shaders[0] || !screen->shaders[1] || !screen->shaders[2])
		return 0;

	screen->program = link_program(screen->shaders[0], screen->shaders[1]);
	if(!screen->program)
		return 0;

	screen->masked_program = link_program(screen->shaders[0], screen->shaders[2]);
	if(!screen->masked_program)
		return 0;

	screen->geometry_loc = glGetUniformLocation(screen->program, "geometry");
	screen->masked_geometry_loc = glGetUniformLocation(screen->masked_program, "geometry");

	loc = glGetUniformLocation(screen->masked_program, "mask");
	if(loc>=0)
	{
		glUseProgram(screen->masked_program);
		glUniform1i(loc, 1);
	}

	glGenBuffers(1, &screen->window_vertex_buffer);
	glBindBuffer(GL_ARRAY_BUFFER, screen->window_vertex_buffer);
	glBufferData(GL_ARRAY_BUFFER, sizeof(window_vertices), window_vertices, GL_STATIC_DRAW);

	stride = 4*sizeof(float);
	glGenVertexArrays(1, &screen->window_vertex_array);
	glBindVertexArray(screen->window_vertex_array);
	glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, stride, NULL);
	glEnableVertexAttribArray(0);
	glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, stride, (void *)(2*sizeof(float)));
	glEnableVertexAttribArray(1);
	glBindVertexArray(0);

	glBindBuffer(GL_ARRAY_BUFFER, 0);

	screen->fb_texture = create_2d_texture();
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB8, screen->width, screen->height, 0, GL_RGB, GL_UNSIGNED_BYTE, NULL);
	glBindTexture(GL_TEXTURE_2D, 0);

	glGenFramebuffers(1, &screen->framebuffer);
	glBindFramebuffer(GL_DRAW_FRAMEBUFFER, screen->framebuffer);
	glFramebufferTexture2D(GL_DRAW_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, screen->fb_texture, 0);
	glDepthMask(GL_FALSE);
	glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);

	screen->root_texture = create_2d_texture();

	return 1;
}

CompositedWindow *find_window(CompositedScreen *screen, Window w)
{
	unsigned i;

	for(i=0; i<screen->nwindows; ++i)
		if(screen->windows[i].window==w)
			return &screen->windows[i];

	return NULL;
}

CompositedWindow *find_window_global(Compositor *compositor, Window w, CompositedScreen **screen)
{
	unsigned i, j;

	for(i=0; i<compositor->nscreens; ++i)
		for(j=0; j<compositor->screens[i].nwindows; ++j)
			if(compositor->screens[i].windows[j].window==w)
			{
				if(screen)
					*screen = &compositor->screens[i];
				return &compositor->screens[i].windows[j];
			}

	if(screen)
		*screen = NULL;
	return NULL;
}

GLXPixmap pixmap_to_glx_pixmap(Compositor *compositor, CompositedScreen *screen, Pixmap pixmap)
{
	int attribs[5];
	unsigned i;

	i = 0;
	attribs[i++] = GLX_TEXTURE_TARGET_EXT;
	attribs[i++] = GLX_TEXTURE_2D_EXT;
	attribs[i++] = GLX_TEXTURE_FORMAT_EXT;
	attribs[i++] = GLX_TEXTURE_FORMAT_RGBA_EXT;
	attribs[i++] = None;

	return glXCreatePixmap(compositor->display, screen->fbconfig, pixmap, attribs);
}

void create_window_pixmap(Compositor *compositor, CompositedScreen *screen, CompositedWindow *window)
{
	if(window->pixmap)
	{
		glXDestroyPixmap(compositor->display, window->glx_pixmap);
		XFreePixmap(compositor->display, window->pixmap);
	}

	window->pixmap = XCompositeNameWindowPixmap(compositor->display, window->window);
	window->glx_pixmap = pixmap_to_glx_pixmap(compositor, screen, window->pixmap);
	window->recreate_pixmap = 0;
}

void update_window_mask(Compositor *compositor, CompositedWindow *window)
{
	Bool bounding_shaped;
	Bool clip_shaped;
	int xi, yi;
	XRectangle *rects;
	int rect_count;
	int rect_order;
	unsigned width;
	unsigned height;
	unsigned char *data;
	int i;
	unsigned y;

	XShapeQueryExtents(compositor->display, window->window, &bounding_shaped, &xi, &yi, &width, &height, &clip_shaped, &xi, &yi, &width, &height);
	window->use_mask = bounding_shaped;
	if(!window->use_mask)
		return;

	rects = XShapeGetRectangles(compositor->display, window->window, ShapeBounding, &rect_count, &rect_order);
	if(!rects)
		return;

	width = window->width+2*window->border;
	height = window->height+2*window->border;
	data = (unsigned char *)malloc(width*height);
	memset(data, 0, width*height);
	for(i=0; i<rect_count; ++i)
	{
		rects[i].x += window->border;
		rects[i].y += window->border;
		if(rects[i].x>=(int)width || rects[i].y>=(int)height)
			continue;

		if(rects[i].x<0)
		{
			if(-rects[i].x>rects[i].width)
				continue;
			rects[i].width += rects[i].x;
			rects[i].x = 0;
		}

		if(rects[i].y<0)
		{
			if(-rects[i].y>rects[i].height)
				continue;
			rects[i].height += rects[i].y;
			rects[i].y = 0;
		}

		if(rects[i].x+rects[i].width>(int)width)
			rects[i].width = width-rects[i].x;
		if(rects[i].y+rects[i].height>(int)height)
			rects[i].height = height-rects[i].y;

		for(y=0; y<rects[i].height; ++y)
		{
			unsigned char *row = data+(rects[i].y+y)*width+rects[i].x;
			memset(row, 255, rects[i].width);
		}
	}

	XFree(rects);

	glBindTexture(GL_TEXTURE_2D, window->mask_texture);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_R8, width, height, 0, GL_RED, GL_UNSIGNED_BYTE, data);

	free(data);

	window->recreate_mask = 0;
}

CompositedWindow *add_window(Compositor *compositor, CompositedScreen *screen, Window w)
{
	CompositedWindow *window;
	XWindowAttributes win_attr;

	if(w==screen->root || w==screen->overlay)
		return NULL;

	if(!XGetWindowAttributes(compositor->display, w, &win_attr))
	{
		printf("XGetWindowAttributes failed; probably the window was already destroyed\n");
		return NULL;
	}
	if(win_attr.class==InputOnly)
		return NULL;

	if(find_window(screen, w))
		return NULL;

	if(screen->nwindows==screen->windows_capacity)
	{
		screen->windows = (CompositedWindow *)realloc(screen->windows, (screen->windows_capacity+1)*sizeof(CompositedWindow));
		++screen->windows_capacity;
	}

	window = &screen->windows[screen->nwindows++];

	window->window = w;
	window->x = win_attr.x;
	window->y = win_attr.y;
	window->width = win_attr.width;
	window->height = win_attr.height;
	window->border = win_attr.border_width;
	window->map_state = win_attr.map_state;

	window->damage = XDamageCreate(compositor->display, window->window, XDamageReportNonEmpty);
	window->pixmap = None;
	window->glx_pixmap = None;

	XCompositeRedirectWindow(compositor->display, window->window, CompositeRedirectManual);
	window->recreate_pixmap = (window->map_state==IsViewable);

	window->texture = create_2d_texture();
	window->mask_texture = create_2d_texture();
	window->use_mask = 0;
	window->recreate_mask = (window->map_state==IsViewable);

	XShapeSelectInput(compositor->display, window->window, ShapeNotifyMask);

	return window;
}

void remove_window(Compositor *compositor, CompositedScreen *screen, CompositedWindow *window, int destroyed)
{
	unsigned i;

	glDeleteTextures(1, &window->texture);
	if(!destroyed)
	{
		XDamageDestroy(compositor->display, window->damage);
		if(window->pixmap)
		{
			glXDestroyPixmap(compositor->display, window->glx_pixmap);
			XFreePixmap(compositor->display, window->pixmap);
		}
		XCompositeUnredirectWindow(compositor->display, window->window, CompositeRedirectManual);
	}

	--screen->nwindows;
	for(i=window-screen->windows; i<screen->nwindows; ++i)
		screen->windows[i] = screen->windows[i+1];
}

CompositedWindow *reorder_window(CompositedScreen *screen, CompositedWindow *window, Window above)
{
	unsigned i, j;
	CompositedWindow hold;

	i = window-screen->windows;
	if((i>0 && screen->windows[i-1].window==above) || (i==0 && !above))
		return window;

	if(above)
	{
		for(j=0; j<screen->nwindows; ++j)
			if(screen->windows[j].window==above)
				break;

		if(j>=screen->nwindows || i==j+1)
			return window;

		if(j<i)
			++j;
	}
	else
		j = 0;

	hold = *window;
	if(i<j)
	{
		for(; i<j; ++i)
			screen->windows[i] = screen->windows[i+1];
	}
	else
	{
		for(; i>j; --i)
			screen->windows[i] = screen->windows[i-1];
	}
	screen->windows[j] = hold;

	return &screen->windows[j];
}

CompositedScreen *find_screen_by_root(Compositor *compositor, Window root)
{
	unsigned i;

	for(i=0; i<compositor->nscreens; ++i)
		if(compositor->screens[i].root==root)
			return &compositor->screens[i];

	return NULL;
}

void create_geometry_correction(CompositedMonitor *monitor)
{
	unsigned buffers[2];
	unsigned stride;

	glGenBuffers(2, buffers);
	monitor->vertex_buffer = buffers[0];
	monitor->index_buffer = buffers[1];

	monitor->tessellation = 50;
	monitor->geometry_data_size = (monitor->tessellation+1)*(monitor->tessellation+1)*3+16;
	monitor->geometry_data = (short *)malloc(monitor->geometry_data_size*sizeof(short));
	stride = 4*sizeof(float);

	glGenVertexArrays(1, &monitor->vertex_array);
	glBindVertexArray(monitor->vertex_array);
	glBindBuffer(GL_ARRAY_BUFFER, monitor->vertex_buffer);
	glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, stride, NULL);
	glEnableVertexAttribArray(0);
	glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, stride, (void *)(2*sizeof(float)));
	glEnableVertexAttribArray(1);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, monitor->index_buffer);
	glBindVertexArray(0);
}

void free_geometry_correction(CompositedMonitor *monitor)
{
	unsigned buffers[2];

	buffers[0] = monitor->vertex_buffer;
	buffers[1] = monitor->index_buffer;
	glDeleteBuffers(2, buffers);
	monitor->vertex_buffer = 0;
	monitor->index_buffer = 0;

	glDeleteVertexArrays(1, &monitor->vertex_array);

	free(monitor->geometry_data);
	monitor->geometry_data = 0;
}

void update_monitor_vertices(CompositedScreen *screen, CompositedMonitor *monitor)
{
	unsigned t;
	unsigned data_size;
	float *vertex_data;
	unsigned short *index_data;
	unsigned x, y;
	unsigned i;
	float aspect;
	float curve_radius;
	float curve_arc;
	float sin_ksv;
	float cos_ksv;
	float distance;
	float eye[3];
	float look[3];

	t = monitor->tessellation;

	data_size = (t+1)*(t+1)*4*sizeof(float);
	vertex_data = (float *)malloc(data_size);

	aspect = (float)monitor->width/monitor->height;

	if(monitor->curvature_depth)
	{
		curve_radius = (monitor->curvature_depth*monitor->curvature_depth+0.25f)/(2.0f*monitor->curvature_depth);
		curve_arc = 2.0f*asin(0.5f/curve_radius);
	}

	sin_ksv = monitor->keystone_vertical/sqrt(1.0f+monitor->keystone_vertical*monitor->keystone_vertical);
	cos_ksv = sqrt(1.0f-sin_ksv*sin_ksv);
	distance = monitor->perspective+sin_ksv*((sin_ksv>0)-monitor->vertical_center)/aspect;
	if(monitor->curvature_depth<0)
		distance += -monitor->curvature_depth;

	eye[0] = 0.0f;
	eye[1] = (monitor->vertical_center-0.5f)/aspect+sin_ksv*distance;
	eye[2] = cos_ksv*distance;

	look[0] = 0.0f;
	look[1] = -sin_ksv;
	look[2] = -cos_ksv;

	for(y=0; y<=t; ++y)
		for(x=0; x<=t; ++x)
		{
			float *v;

			v = vertex_data+(y*(t+1)+x)*3;
			v[0] = (float)x/t-0.5f;
			v[1] = ((float)y/t-0.5f)/aspect;
			v[2] = 0;
			if(monitor->curvature_depth)
			{
				if(monitor->curvature_type==CYLINDRICAL)
				{
					v[2] = (1.0f-cos(v[0]*curve_arc))*curve_radius-monitor->curvature_depth;
					v[0] = sin(v[0]*curve_arc)*curve_radius;
				}
				else if(monitor->curvature_type==SPHERICAL)
				{
					float r;

					v[0] = tan(v[0]*curve_arc)*curve_radius;
					v[1] = tan(v[1]*curve_arc)*curve_radius;
					r = sqrt(v[0]*v[0]+v[1]*v[1]+curve_radius*curve_radius)/fabs(curve_radius);
					v[0] /= r;
					v[1] /= r;
					v[2] = curve_radius-curve_radius/r-monitor->curvature_depth;
				}
			}
		}

	monitor->geometry_data[0] = t;
	for(i=0; i<3; ++i)
		monitor->geometry_data[1+i] = eye[i]*4096;
	for(y=0; y<2; ++y)
		for(x=0; x<2; ++x)
		{
			i = 1+(1+y*2+x)*3;
			monitor->geometry_data[i] = ((x-0.5f)*distance/monitor->perspective)*4096;
			monitor->geometry_data[i+1] = (eye[1]+look[1]*distance-(y-monitor->vertical_center)*look[2]*distance/monitor->perspective/aspect)*4096;
			monitor->geometry_data[i+2] = (eye[2]+look[2]*distance+(y-monitor->vertical_center)*look[1]*distance/monitor->perspective/aspect)*4096;
		}
	for(i=0; i<(t+1)*(t+1)*3; ++i)
		monitor->geometry_data[16+i] = vertex_data[i]*4096;

	for(y=t; y<=t; --y)
		for(x=t; x<=t; --x)
		{
			float *v;
			float px, py, pz;
			float scale;

			v = vertex_data+(y*(t+1)+x)*3;
			px = v[0]-eye[0];
			py = (v[0]-eye[0])*look[0] - (v[1]-eye[1])*look[2] + (v[2]-eye[2])*look[1];
			pz = (v[0]-eye[0])*look[0] + (v[1]-eye[1])*look[1] + (v[2]-eye[2])*look[2];
			scale = monitor->perspective/pz;

			v = vertex_data+(y*(t+1)+x)*4;
			v[0] = px*scale+0.5f;
			v[1] = py*aspect*scale+monitor->vertical_center;
			v[2] = (monitor->x+(float)x*monitor->width/t)/screen->width;
			v[3] = 1.0f-(monitor->y+(1.0f-(float)y/t)*monitor->height)/screen->height;
		}

	glBindBuffer(GL_ARRAY_BUFFER, monitor->vertex_buffer);
	glBufferData(GL_ARRAY_BUFFER, data_size, vertex_data, GL_STATIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	free(vertex_data);

	monitor->nelements = t*((t+1)*2+1)-1;
	data_size = monitor->nelements*sizeof(unsigned short);
	index_data = (unsigned short *)malloc(data_size);
	i = 0;
	for(y=0; y<t; ++y)
	{
		if(y>0)
			index_data[i++] = 0xFFFF;
		for(x=0; x<=t; ++x)
		{
			index_data[i++] = (y+1)*(t+1)+x;
			index_data[i++] = y*(t+1)+x;
		}
	}
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, monitor->index_buffer);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, data_size, index_data, GL_STATIC_DRAW);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
	free(index_data);
}

int initialize_monitor(Compositor *compositor, CompositedScreen *screen, XRRScreenResources *xrr_res, unsigned index)
{
	CompositedMonitor *monitor;
	XRROutputInfo *output;
	int namelen;
	XRRCrtcInfo *crtc;

	monitor = &screen->monitors[index];

	output = XRRGetOutputInfo(compositor->display, xrr_res, xrr_res->outputs[index]);
	namelen = strlen(output->name);
	monitor->name = (char *)malloc(namelen+1);
	strcpy(monitor->name, output->name);
	monitor->enabled = !!output->crtc;
	if(!monitor->enabled)
	{
		XRRFreeOutputInfo(output);
		monitor->vertex_buffer = 0;
		monitor->index_buffer = 0;
		monitor->vertex_array = 0;
		monitor->geometry_data = NULL;
		return 1;
	}

	crtc = XRRGetCrtcInfo(compositor->display, xrr_res, output->crtc);
	monitor->x = crtc->x;
	monitor->y = crtc->y;
	monitor->width = crtc->width;
	monitor->height = crtc->height;
	XRRFreeCrtcInfo(crtc);
	XRRFreeOutputInfo(output);

	monitor->keystone_vertical = 0.0f;
	monitor->curvature_type = CYLINDRICAL;
	monitor->curvature_depth = 0.0f;
	monitor->vertical_center = 0.5f;
	monitor->perspective = 1.0f;

	create_geometry_correction(monitor);
	update_monitor_vertices(screen, monitor);

	return 1;
}

CompositedMonitor *find_monitor_by_name(CompositedScreen *screen, char *name)
{
	unsigned i;

	for(i=0; i<screen->nmonitors; ++i)
		if(!strcmp(screen->monitors[i].name, name))
			return &screen->monitors[i];

	return NULL;
}

CompositedMonitor *find_monitor_by_name_global(Compositor *compositor, char *name, CompositedScreen **screen)
{
	unsigned i, j;

	for(i=0; i<compositor->nscreens; ++i)
		for(j=0; j<compositor->screens[i].nmonitors; ++j)
			if(!strcmp(compositor->screens[i].monitors[j].name, name))
			{
				if(screen)
					*screen = &compositor->screens[i];
				return &compositor->screens[i].monitors[j];
			}

	if(screen)
		*screen = NULL;
	return NULL;
}

void update_monitors(Compositor *compositor, CompositedScreen *screen)
{
	XRRScreenResources *xrr_res;
	int i;

	xrr_res = XRRGetScreenResources(compositor->display, screen->root);
	for(i=0; i<xrr_res->noutput; ++i)
	{
		XRROutputInfo *output;
		CompositedMonitor *monitor;
		int was_enabled;
		XRRCrtcInfo *crtc;

		output = XRRGetOutputInfo(compositor->display, xrr_res, xrr_res->outputs[i]);
		monitor = find_monitor_by_name(screen, output->name);
		if(!monitor)
			continue;

		was_enabled = monitor->enabled;
		monitor->enabled = !!output->crtc;
		if(!monitor->enabled)
		{
			XRRFreeOutputInfo(output);
			if(was_enabled)
				free_geometry_correction(monitor);
			continue;
		}

		crtc = XRRGetCrtcInfo(compositor->display, xrr_res, output->crtc);
		monitor->x = crtc->x;
		monitor->y = crtc->y;
		monitor->width = crtc->width;
		monitor->height = crtc->height;
		XRRFreeCrtcInfo(crtc);
		XRRFreeOutputInfo(output);

		if(!was_enabled)
			create_geometry_correction(monitor);

		update_monitor_vertices(screen, monitor);
	}
	XRRFreeScreenResources(xrr_res);
}

void update_geometry_correction(Compositor *compositor, CompositedScreen *screen)
{
	Atom prop_type;
	int prop_format;
	unsigned long overflow;
	unsigned long names_length;
	char *names;
	unsigned long values_length;
	short *values;
	char *name_ptr;
	unsigned i;

	XGetWindowProperty(compositor->display, screen->root, compositor->monitors_atom, 0, 64, False, XA_STRING,
		&prop_type, &prop_format, &names_length, &overflow, (unsigned char **)&names);
	if(prop_type!=XA_STRING || prop_format!=8)
		return;

	XGetWindowProperty(compositor->display, screen->root, compositor->correction_atom, 0, 64, False, XA_INTEGER,
		&prop_type, &prop_format, &values_length, &overflow, (unsigned char **)&values);
	if(prop_type!=XA_INTEGER || prop_format!=16)
	{
		XFree(names);
		return;
	}

	use_gl(compositor, screen);

	name_ptr = names;
	for(i=0; i*5+4<values_length; ++i)
	{
		CompositedMonitor *monitor;

		if(name_ptr>=names+names_length)
			break;

		monitor = find_monitor_by_name(screen, name_ptr);
		if(monitor)
		{
			monitor->keystone_vertical = values[i*5]/4096.0f;
			monitor->curvature_type = values[i*5+1];
			monitor->curvature_depth = values[i*5+2]/4096.0f;
			monitor->vertical_center = values[i*5+3]/4096.0f;
			monitor->perspective = values[i*5+4]/4096.0f;

			if(monitor->enabled)
				update_monitor_vertices(screen, monitor);
		}

		name_ptr += strlen(name_ptr)+1;
	}

	XFree(names);
	XFree(values);
}

void update_root_pixmap(Compositor *compositor, CompositedScreen *screen)
{
	Atom prop_type;
	int prop_format;
	unsigned long overflow;
	unsigned long length;
	long *pixmap;
	Window root;
	int x, y;
	unsigned width;
	unsigned height;
	unsigned border;
	unsigned depth;

	use_gl(compositor, screen);

	if(screen->root_glx_pixmap)
	{
		glXDestroyPixmap(compositor->display, screen->root_glx_pixmap);
		screen->root_glx_pixmap = None;
	}

	XGetWindowProperty(compositor->display, screen->root, compositor->root_pmap_atom, 0, 1, False, XA_PIXMAP,
		&prop_type, &prop_format, &length, &overflow, (unsigned char **)&pixmap);
	if(prop_type!=XA_PIXMAP || prop_format!=32)
	{
		screen->root_pixmap = None;
		return;
	}

	screen->root_pixmap = pixmap[0];
	if(XGetGeometry(compositor->display, screen->root_pixmap, &root, &x, &y, &width, &height, &border, &depth))
		screen->root_glx_pixmap = pixmap_to_glx_pixmap(compositor, screen, screen->root_pixmap);
	else
		screen->root_pixmap = None;

	XFree(pixmap);
}

int initialize_screen(Compositor *compositor, unsigned number)
{
	CompositedScreen *screen;
	const char *extensions;
	Window dummy_root;
	int x, y;
	unsigned border;
	unsigned depth;
	XRRScreenResources *xrr_res;
	Window dummy_parent;
	Window *children;
	unsigned nchildren;
	unsigned i;

	screen = &compositor->screens[number];
	screen->number = number;

	extensions = glXQueryExtensionsString(compositor->display, screen->number);
	if(!strstr(extensions, "GLX_ARB_create_context"))
		return with_error("GLX_ARB_create_context is required");
	if(!strstr(extensions, "GLX_EXT_texture_from_pixmap"))
		return with_error("GLX_EXT_texture_from_pixmap is required");

	screen->root = RootWindow(compositor->display, screen->number);
	XSelectInput(compositor->display, screen->root, SubstructureNotifyMask|PropertyChangeMask);
	screen->overlay = XCompositeGetOverlayWindow(compositor->display, screen->root);
	XSelectInput(compositor->display, screen->overlay, StructureNotifyMask);
	XGetGeometry(compositor->display, screen->overlay, &dummy_root, &x, &y, &screen->width, &screen->height, &border, &depth);
	XShapeCombineRectangles(compositor->display, screen->overlay, ShapeInput, 0, 0, NULL, 0, ShapeSet, Unsorted);

	if(!initialize_gl(compositor, screen))
		return 0;

	use_gl(compositor, screen);

	if(!create_gl_resources(screen))
		return 0;

	xrr_res = XRRGetScreenResources(compositor->display, screen->root);
	screen->nmonitors = xrr_res->noutput;
	screen->monitors = (CompositedMonitor *)malloc(screen->nmonitors*sizeof(CompositedMonitor));
	for(i=0; i<screen->nmonitors; ++i)
		if(!initialize_monitor(compositor, screen, xrr_res, i))
			return 0;
	XRRFreeScreenResources(xrr_res);
	XRRSelectInput(compositor->display, screen->root, RRScreenChangeNotifyMask);

	screen->root_pixmap = None;
	screen->root_glx_pixmap = None;

	update_geometry_correction(compositor, screen);
	update_root_pixmap(compositor, screen);

	XQueryTree(compositor->display, screen->root, &dummy_root, &dummy_parent, &children, &nchildren);

	screen->windows = (CompositedWindow *)malloc(nchildren*sizeof(CompositedWindow));
	screen->nwindows = 0;
	screen->windows_capacity = nchildren;

	for(i=0; i<nchildren; ++i)
		add_window(compositor, screen, children[i]);

	XFree(children);

	screen->dirty = 1;

	return 1;
}

int initialize_compositor(Compositor *compositor)
{
	int event_base;
	int error_base;
	int major_ver;
	int minor_ver;
	unsigned i;

	compositor->display = XOpenDisplay(NULL);
	if(!compositor->display)
		return with_error("Could not open X display");

	XSetErrorHandler(&x_error_handler);

	if(!XCompositeQueryExtension(compositor->display, &event_base, &error_base))
		return with_error("XComposite is required but was not found");
	else if(!XCompositeQueryVersion(compositor->display, &major_ver, &minor_ver))
		return with_error("Cannot determine XComposite version");
	else if(major_ver==0 && minor_ver<3)
		return with_error("XComposite 0.3 or later is required");

	if(!glXQueryExtension(compositor->display, &event_base, &error_base))
		return with_error("GLX is required but was not found");
	else if(!glXQueryVersion(compositor->display, &major_ver, &minor_ver))
		return with_error("Cannot determine GLX version");
	else if(major_ver<1 || (major_ver==1 && minor_ver<4))
		return with_error("GLX 1.4 or later is required");

	if(!XDamageQueryExtension(compositor->display, &compositor->damage_event, &error_base))
		return with_error("XDamage is required but was not found");
	else if(!XDamageQueryVersion(compositor->display, &major_ver, &minor_ver))
		return with_error("Cannot determine XDamage version");
	else if(major_ver<1)
		return with_error("XDamage 1.0 or later is required");

	if(!XShapeQueryExtension(compositor->display, &compositor->shape_event, &error_base))
		return with_error("XShape is required but was not found");
	else if(!XShapeQueryVersion(compositor->display, &major_ver, &minor_ver))
		return with_error("Cannot determine XShape version");
	else if(major_ver<1 || (major_ver==1 && minor_ver<1))
		return with_error("XShape 1.1 or later is required");

	if(!XRRQueryExtension(compositor->display, &compositor->randr_event, &error_base))
		return with_error("XRandR is required but was not found");
	else if(!XRRQueryVersion(compositor->display, &major_ver, &minor_ver))
		return with_error("Cannot determine XRandR version");
	else if(major_ver<1 || (major_ver==1 && minor_ver<2))
		return with_error("XRandR 1.2 or later is required");

	compositor->glXCreateContextAttribsARB = (PFNGLXCREATECONTEXTATTRIBSARBPROC)glXGetProcAddress((unsigned char *)"glXCreateContextAttribsARB");
	compositor->glXBindTexImageEXT = (PFNGLXBINDTEXIMAGEEXTPROC)glXGetProcAddress((unsigned char *)"glXBindTexImageEXT");
	compositor->glXReleaseTexImageEXT = (PFNGLXRELEASETEXIMAGEEXTPROC)glXGetProcAddress((unsigned char *)"glXReleaseTexImageEXT");

	compositor->root_pmap_atom = XInternAtom(compositor->display, "_XROOTPMAP_ID", False);
	compositor->correction_atom = XInternAtom(compositor->display, "_MSP_GEOMETRY_CORRECTION", False);
	compositor->monitors_atom = XInternAtom(compositor->display, "_MSP_GEOMETRY_CORRECTION_MONITORS", False);
	compositor->geometry_data_atom = XInternAtom(compositor->display, "_MSP_GEOMETRY_CORRECTION_DATA", False);

	compositor->nscreens = ScreenCount(compositor->display);
	compositor->screens = (CompositedScreen *)malloc(compositor->nscreens*sizeof(CompositedScreen));
	for(i=0; i<compositor->nscreens; ++i)
		if(!initialize_screen(compositor, i))
			return 0;

	compositor->selection_window = XCreateWindow(compositor->display, compositor->screens[0].root, 0, 0, 1, 1, 0, CopyFromParent, InputOutput, CopyFromParent, 0, NULL);
	XSetSelectionOwner(compositor->display, compositor->geometry_data_atom, compositor->selection_window, CurrentTime);

	compositor->dirty = 1;

	return 1;
}

void shutdown_screen(Compositor *compositor, CompositedScreen *screen)
{
	unsigned i;

	use_gl(compositor, screen);

	for(i=0; i<screen->nwindows; ++i)
	{
		glDeleteTextures(1, &screen->windows[i].texture);
		glDeleteTextures(1, &screen->windows[i].mask_texture);
		if(screen->windows[i].pixmap)
		{
			glXDestroyPixmap(compositor->display, screen->windows[i].glx_pixmap);
			XFreePixmap(compositor->display, screen->windows[i].pixmap);
			XDamageDestroy(compositor->display, screen->windows[i].damage);
		}
	}

	for(i=0; i<screen->nmonitors; ++i)
	{
		free(screen->monitors[i].name);
		if(screen->monitors[i].enabled)
			free_geometry_correction(&screen->monitors[i]);
	}

	glDeleteTextures(1, &screen->root_texture);
	glXDestroyPixmap(compositor->display, screen->root_glx_pixmap);

	glDeleteBuffers(1, &screen->window_vertex_buffer);
	glDeleteVertexArrays(1, &screen->window_vertex_array);
	glDeleteFramebuffers(1, &screen->framebuffer);
	glDeleteTextures(1, &screen->fb_texture);
	glDeleteProgram(screen->program);
	glDeleteProgram(screen->masked_program);
	for(i=0; i<3; ++i)
		glDeleteShader(screen->shaders[i]);

	glXMakeContextCurrent(compositor->display, None, None, NULL);
	glXDestroyContext(compositor->display, screen->glx_context);
	glXDestroyWindow(compositor->display, screen->glx_window);
	XDestroyWindow(compositor->display, screen->render_window);

	XCompositeReleaseOverlayWindow(compositor->display, screen->overlay);

	free(screen->windows);
	free(screen->monitors);
}

void shutdown_compositor(Compositor *compositor)
{
	unsigned i;

	for(i=0; i<compositor->nscreens; ++i)
		shutdown_screen(compositor, &compositor->screens[i]);
	free(compositor->screens);

	XDestroyWindow(compositor->display, compositor->selection_window);

	XCloseDisplay(compositor->display);
}

void mark_dirty(Compositor *compositor, CompositedScreen *screen)
{
	compositor->dirty = 1;
	screen->dirty = 1;
}

void process_create_window_event(Compositor *compositor, XCreateWindowEvent *event)
{
	CompositedScreen *screen;

	if((screen = find_screen_by_root(compositor, event->parent)))
		add_window(compositor, screen, event->window);
}

void process_destroy_window_event(Compositor *compositor, XDestroyWindowEvent *event)
{
	CompositedScreen *screen;
	CompositedWindow *window;

	if((screen = find_screen_by_root(compositor, event->event)))
		if((window = find_window(screen, event->window)))
		{
			use_gl(compositor, screen);
			remove_window(compositor, screen, window, 1);
		}
}

void process_map_event(Compositor *compositor, XMapEvent *event)
{
	CompositedScreen *screen;
	CompositedWindow *window;

	screen = find_screen_by_root(compositor, event->event);
	if(!screen)
		return;

	window = find_window(screen, event->window);
	if(!window)
		return;

	window->map_state = IsViewable;
	window->recreate_pixmap = 1;
	window->recreate_mask = 1;

	mark_dirty(compositor, screen);
}

void process_unmap_event(Compositor *compositor, XUnmapEvent *event)
{
	CompositedScreen *screen;
	CompositedWindow *window;

	screen = find_screen_by_root(compositor, event->event);
	if(!screen)
		return;

	window = find_window(screen, event->window);
	if(window)
		window->map_state = IsUnviewable;

	mark_dirty(compositor, screen);
}

void process_reparent_event(Compositor *compositor, XReparentEvent *event)
{
	CompositedScreen *screen;
	CompositedWindow *window;

	screen = find_screen_by_root(compositor, event->event);
	if(!screen)
		return;

	if(event->parent==screen->root)
		window = add_window(compositor, screen, event->window);
	else
	{
		window = find_window(screen, event->window);
		if(!window)
			return;

		remove_window(compositor, screen, window, 0);
	}

	if(window && window->map_state==IsViewable)
		mark_dirty(compositor, screen);
}

void process_configure_event(Compositor *compositor, XConfigureEvent *event)
{
	CompositedScreen *screen;
	CompositedWindow *window;

	screen = find_screen_by_root(compositor, event->event);
	if(!screen)
		return;

	if(event->window==screen->overlay)
	{
		screen->width = event->width;
		screen->height = event->height;

		XResizeWindow(compositor->display, screen->render_window, screen->width, screen->height);
		use_gl(compositor, screen);
		glBindTexture(GL_TEXTURE_2D, screen->fb_texture);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB8, screen->width, screen->height, 0, GL_RGB, GL_UNSIGNED_BYTE, NULL);
		glBindTexture(GL_TEXTURE_2D, 0);
		return;
	}

	window = find_window(screen, event->window);
	if(!window)
		return;

	window->x = event->x;
	window->y = event->y;
	if((unsigned)event->width!=window->width || (unsigned)event->height!=window->height || (unsigned)event->border_width!=window->border)
	{
		window->width = event->width;
		window->height = event->height;
		window->border = event->border_width;
		window->recreate_pixmap = 1;
	}
	window = reorder_window(screen, window, event->above);

	if(window->map_state==IsViewable)
		mark_dirty(compositor, screen);
}

void process_property_event(Compositor *compositor, XPropertyEvent *event)
{
	CompositedScreen *screen;

	screen = find_screen_by_root(compositor, event->window);
	if(!screen)
		return;

	if(event->atom==compositor->correction_atom)
	{
		update_geometry_correction(compositor, screen);
		mark_dirty(compositor, screen);
	}
	else if(event->atom==compositor->root_pmap_atom)
	{
		update_root_pixmap(compositor, screen);
		mark_dirty(compositor, screen);
	}
}

void process_selection_request_event(Compositor *compositor, XSelectionRequestEvent *event)
{
	Atom prop_type;
	int prop_format;
	unsigned long overflow;
	unsigned long names_length;
	XSelectionEvent notify;

	if(event->selection==compositor->geometry_data_atom)
	{
		char *monitor_name;
		CompositedMonitor *monitor;

		XGetWindowProperty(compositor->display, event->requestor, event->property, 0, 64, False, XA_STRING,
			&prop_type, &prop_format, &names_length, &overflow, (unsigned char **)&monitor_name);
		if(prop_type!=XA_STRING || prop_format!=8)
			return;

		monitor = find_monitor_by_name_global(compositor, monitor_name, NULL);
		if(monitor && monitor->enabled)
			XChangeProperty(compositor->display, event->requestor, event->property, XA_INTEGER, 16, PropModeReplace, (unsigned char *)monitor->geometry_data, monitor->geometry_data_size);

		notify.type = SelectionNotify;
		notify.requestor = event->requestor;
		notify.selection = event->selection;
		notify.target = event->target;
		notify.property = (monitor ? event->property : None);
		notify.time = event->time;
		XSendEvent(compositor->display, event->requestor, False, 0, (XEvent *)&notify);
	}
}

void process_damage_event(Compositor *compositor, XDamageNotifyEvent *event)
{
	CompositedScreen *screen;
	CompositedWindow *window;

	window = find_window_global(compositor, event->drawable, &screen);
	if(window && window->map_state==IsViewable)
		mark_dirty(compositor, screen);
}

void process_shape_event(Compositor *compositor, XShapeEvent *event)
{
	CompositedScreen *screen;
	CompositedWindow *window;

	if(event->kind!=ShapeBounding)
		return;

	window = find_window_global(compositor, event->window, &screen);
	if(window && window->map_state==IsViewable)
	{
		window->recreate_mask = 1;
		mark_dirty(compositor, screen);
	}
}

void process_randr_event(Compositor *compositor, XRRScreenChangeNotifyEvent *event)
{
	CompositedScreen *screen;

	if((screen = find_screen_by_root(compositor, event->root)))
		update_monitors(compositor, screen);
}

void process_events(Compositor *compositor)
{
	int pending;
	XEvent event;

	pending = 0;
	while((pending || !compositor->dirty) && !terminate_requested)
	{
		if(!pending)
			pending = XPending(compositor->display);

		XNextEvent(compositor->display, &event);
		if(pending)
			--pending;

		switch(event.type)
		{
		case CreateNotify:
			process_create_window_event(compositor, &event.xcreatewindow);
			break;
		case DestroyNotify:
			process_destroy_window_event(compositor, &event.xdestroywindow);
			break;
		case MapNotify:
			process_map_event(compositor, &event.xmap);
			break;
		case UnmapNotify:
			process_unmap_event(compositor, &event.xunmap);
			break;
		case ReparentNotify:
			process_reparent_event(compositor, &event.xreparent);
			break;
		case ConfigureNotify:
			process_configure_event(compositor, &event.xconfigure);
			break;
		case PropertyNotify:
			process_property_event(compositor, &event.xproperty);
			break;
		case SelectionRequest:
			process_selection_request_event(compositor, &event.xselectionrequest);
			break;
		default:
			if(event.type==compositor->damage_event+XDamageNotify)
				process_damage_event(compositor, (XDamageNotifyEvent *)&event);
			else if(event.type==compositor->shape_event+ShapeNotify)
				process_shape_event(compositor, (XShapeEvent *)&event);
			else if(event.type==compositor->randr_event+RRScreenChangeNotify)
				process_randr_event(compositor, (XRRScreenChangeNotifyEvent *)&event);
		}
	}
}

void refresh_screen(Compositor *compositor, CompositedScreen *screen)
{
	unsigned i;
	int use_mask;

	for(i=0; i<screen->nwindows; ++i)
		if(screen->windows[i].map_state==IsViewable)
			XDamageSubtract(compositor->display, screen->windows[i].damage, None, None);
	glXWaitX();

	use_gl(compositor, screen);

	glViewport(0, 0, screen->width, screen->height);
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glBindFramebuffer(GL_DRAW_FRAMEBUFFER, screen->framebuffer);

	use_mask = -1;
	glBindVertexArray(screen->window_vertex_array);

	if(screen->root_pixmap)
	{
		use_mask = 0;
		glUseProgram(screen->program);
		glBindTexture(GL_TEXTURE_2D, screen->root_texture);
		compositor->glXBindTexImageEXT(compositor->display, screen->root_glx_pixmap, GLX_FRONT_LEFT_EXT, NULL);
		glUniform4f(screen->geometry_loc, 0.0f, 0.0f, 1.0f, 1.0f);
		glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
		compositor->glXReleaseTexImageEXT(compositor->display, screen->root_glx_pixmap, GLX_FRONT_LEFT_EXT);
	}
	else
		glClear(GL_COLOR_BUFFER_BIT);

	for(i=0; i<screen->nwindows; ++i)
	{
		CompositedWindow *window;

		window = &screen->windows[i];
		if(window->map_state!=IsViewable)
			continue;

		if(window->use_mask!=use_mask)
		{
			use_mask = window->use_mask;
			glUseProgram(use_mask ? screen->masked_program : screen->program);
		}

		if(window->use_mask)
		{
			glActiveTexture(GL_TEXTURE1);
			glBindTexture(GL_TEXTURE_2D, window->mask_texture);
			glActiveTexture(GL_TEXTURE0);
		}

		if(window->recreate_pixmap)
			create_window_pixmap(compositor, screen, window);
		if(window->recreate_mask)
			update_window_mask(compositor, window);

		glBindTexture(GL_TEXTURE_2D, window->texture);
		compositor->glXBindTexImageEXT(compositor->display, window->glx_pixmap, GLX_FRONT_LEFT_EXT, NULL);
		glUniform4f((use_mask ? screen->masked_geometry_loc : screen->geometry_loc),
			(float)window->x/screen->width, 1.0f-(float)(window->y+window->height)/screen->height,
			(float)(window->width+2*window->border)/screen->width, (float)(window->height+2*window->border)/screen->height);
		glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
		compositor->glXReleaseTexImageEXT(compositor->display, window->glx_pixmap, GLX_FRONT_LEFT_EXT);
	}

	glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);
	glClear(GL_COLOR_BUFFER_BIT);
	glBindTexture(GL_TEXTURE_2D, screen->fb_texture);
	glEnable(GL_PRIMITIVE_RESTART);
	glPrimitiveRestartIndex(0xFFFF);
	if(use_mask)
		glUseProgram(screen->program);

	for(i=0; i<screen->nmonitors; ++i)
	{
		CompositedMonitor *monitor = &screen->monitors[i];
		if(!monitor->enabled)
			continue;

		glUniform4f(screen->geometry_loc,
			(float)monitor->x/screen->width, 1.0f-(float)(monitor->y+monitor->height)/screen->height,
			(float)monitor->width/screen->width, (float)monitor->height/screen->height);

		glBindVertexArray(monitor->vertex_array);
		glDrawElements(GL_TRIANGLE_STRIP, monitor->nelements, GL_UNSIGNED_SHORT, NULL);
	}

	glBindVertexArray(0);

	glXSwapBuffers(compositor->display, screen->glx_window);

	screen->dirty = 0;
}

void refresh_all_screens(Compositor *compositor)
{
	unsigned i;

	for(i=0; i<compositor->nscreens; ++i)
	{
		CompositedScreen *screen = &compositor->screens[i];
		if(screen->dirty)
			refresh_screen(compositor, screen);
	}

	compositor->dirty = 0;
}

void sighandler(int sig)
{
	terminate_requested = 1;
	(void)sig;
}

int main()
{
	Compositor compositor;

	signal(SIGINT, &sighandler);
	signal(SIGTERM, &sighandler);

	if(!initialize_compositor(&compositor))
		return 1;

	while(!terminate_requested)
	{
		process_events(&compositor);
		if(compositor.dirty && !terminate_requested)
			refresh_all_screens(&compositor);
	}

	shutdown_compositor(&compositor);

	return 0;
}
